Package: ids
Title: Generate Random Identifiers
Version: 1.0.1
Authors@R: person("Rich", "FitzJohn", role = c("aut", "cre"),
    email = "rich.fitzjohn@gmail.com")
Description: Generate random or human readable and pronounceable identifiers.
License: MIT + file LICENSE
URL: https://github.com/richfitz/ids
BugReports: https://github.com/richfitz/ids/issues
Imports: openssl, uuid
Suggests: knitr, rcorpora, rmarkdown, testthat
RoxygenNote: 6.0.1
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2017-05-31 08:43:42 UTC; rich
Author: Rich FitzJohn [aut, cre]
Maintainer: Rich FitzJohn <rich.fitzjohn@gmail.com>
Repository: CRAN
Date/Publication: 2017-05-31 08:49:59 UTC
Built: R 4.6.0; ; 2026-04-26 02:30:21 UTC; unix

# Generated by roxygen2: do not edit by hand

export(adjective_animal)
export(ids)
export(int_to_proquint)
export(int_to_proquint_word)
export(proquint)
export(proquint_to_int)
export(proquint_word_to_int)
export(random_id)
export(sentence)
export(uuid)
importFrom(uuid,UUIDgenerate)

# ids

[![Project Status: Active - The project has reached a stable, usable state and is being actively developed.](http://www.repostatus.org/badges/latest/active.svg)](http://www.repostatus.org/#active)
[![Linux Build Status](https://travis-ci.org/richfitz/ids.svg?branch=master)](https://travis-ci.org/richfitz/ids)
[![codecov.io](https://codecov.io/github/richfitz/ids/coverage.svg?branch=master)](https://codecov.io/github/richfitz/ids?branch=master)
[![](https://www.r-pkg.org/badges/version/ids)](https://cran.r-project.org/package=ids)

Generate random identifiers in a number of styles:

* random ids of any number of bytes, such as `31f6d556fe2b303c`
* UUIDs using the `uuid` package, such as `4f0efabf-0375-4a08-89ea-b8f162f07c44`
* human readable identifiers in the style `<adjective>_<animal>` (following [gfycat.com](http://gfycat.com)), such as `misanthropic_lungfish`
* human readable identifiers in the style of a sentence (following [Asana](https://blog.asana.com/2011/09/6-sad-squid-snuggle-softly), such as `33_enormous_chinchillas_tumbling_elegantly`
* [proquint](https://arxiv.org/html/0901.4016) ("PRO-nouncable QUINT-uplet") idenfiers such as `dizuz-soboz` (which can be convereted to an integer such as 40,2638,895)

These can all be tweaked with options for length, words that are used, and the case that joins words.  There is a function `ids` for creating your own human readable identifiers.

## Installation

Install this package from CRAN

```r
install.packages("ids")
```

or install the development version with

```r
remotes::install_github("richfitz/ids", upgrade = FALSE)
```

## ---
## title: "ids"
## author: "Rich FitzJohn"
## date: "`r Sys.Date()`"
## output: rmarkdown::html_vignette
## vignette: >
##   %\VignetteIndexEntry{ids}
##   %\VignetteEngine{knitr::rmarkdown}
##   %\VignetteEncoding{UTF-8}
## ---

##+ echo = FALSE, results = "hide"
knitr::opts_chunk$set(error = FALSE)
human_no <- function(x) {
  s <- log10(floor(x + 1))
  p <- c(0, thousand = 3, million = 6, billion = 9, trillion = 12)
  i <- s > p
  j <- max(which(i))
  str <- names(p)[j]
  if (nzchar(str)) {
    paste(signif(x / 10^p[[j]], 3), str)
  } else {
    as.character(x)
  }
}
set.seed(1)

## The `ids` package provides randomly generated ids in a number of
## different forms with different readability and sizes.

## ## Random bytes

## The `random_id` function generates random identifiers by generating
## `bytes` random bytes and converting to hexadecimal (so each byte
## becomes a pair of characters).  Rather than use R's random number
## stream we use the `openssl` package here.
ids::random_id()

## All `ids` functions take `n` as the first argument to be the number
## of identifiers generated:
ids::random_id(5)

## The default here is 16 bytes, each of which has 256 values (so
## 256^16 = 2^128 = 3.4e38 combinations).  You can make these larger or
## smaller with the `bytes` argument:
ids::random_id(5, 8)

## If `NULL` is provided as `n`, then a generating function is
## returned (all ids functions do this):
f <- ids::random_id(NULL, 8)
f

## This function sets all arguments except for `n`
f()
f(4)

## ## UUIDs

## The above look a lot like UUIDs but they are not actually UUIDs.
## The `uuid` package provides real UUIDs generated with libuuid, and
## the `ids::uuid` function provides an interface to that:
ids::uuid()

## As above, generate more than one UUID:
ids::uuid(4)

## Generate time-based UUIDs:
ids::uuid(4, use_time = TRUE)

## and optionally drop the hyphens:
ids::uuid(5, drop_hyphens = TRUE)

## ## Adjective animal

## Generate (somewhat) human readable identifiers by combining one or
## more adjectives with an animal name.
ids::adjective_animal()

## The list of adjectives and animals comes from
## [gfycat.com](http://gfycat.com), via
## https://github.com/a-type/adjective-adjective-animal

## Generate more than one identifier:
ids::adjective_animal(4)

## Use more than one adjective for very long idenfiers
ids::adjective_animal(4, 3)

##+ echo = FALSE, results = "hide"
n1 <- length(ids:::gfycat_animals)
n2 <- length(ids:::gfycat_adjectives)

## There are `r n1` animal names and `r n2` adjectives so each one you
## add increases the idenfier space by a factor of `r n2`.  So for 1,
## 2, and 3 adjectives there are about `r human_no(n1 * n2)`,
## `r human_no(n1 * n2^2)` and `r human_no(n1 * n2^3)` possible combinations.

## This is a much smaller space than the random identifiers above, but
## these are more readable and memorable.

## Note that here, the random nunbers are coming from R's random
## number stream so are affected by `set.seed()`.

## Because some of the animal and adjective names are very long
## (e.g. a _quasiextraterritorial hexakosioihexekontahexaphobic
## queenalexandrasbirdwingbutterfly_), in order to generate more
## readable/memorable identifiers it may be useful to restrict the
## length.  Pass `max_len` in to do this.
ids::adjective_animal(4, max_len = 6)

## A vector of length 2 here can be used to apply to the adjectives
## and animal respectively:
ids::adjective_animal(20, max_len = c(5, Inf))

## Note that this decreases the pool size and so increases the chance
## of collisions.

## In addition to snake_case, the default, the punctuation between
## words can be changed to:
##
## kebab-case:
ids::adjective_animal(1, 2, style = "kebab")

## dot.case:
ids::adjective_animal(1, 2, style = "dot")

## camel-case:
ids::adjective_animal(1, 2, style = "camel")

## PascalCase:
ids::adjective_animal(1, 2, style = "pascal")

## CONSTANT_CASE (aka SHOUTY_CASE)
ids::adjective_animal(1, 2, style = "constant")

## or with spaces, lower case:
ids::adjective_animal(1, 2, style = "lower")

## UPPER CASE
ids::adjective_animal(1, 2, style = "upper")

## Sentence case
ids::adjective_animal(1, 2, style = "sentence")

## Title Case
ids::adjective_animal(1, 2, style = "title")

## Again, pass `n = NULL` here to create a generating function:
aa3 <- ids::adjective_animal(NULL, 3, style = "kebab", max_len = c(6, 8))

## ...which can be used to generate ids on demand.
aa3()
aa3(4)

## ## Random sentences

## The `sentence` function creates a sentence style identifier.  This
## uses the approach described by Asana on [their
## blog](https://blog.asana.com/2011/09/6-sad-squid-snuggle-softly).
## This approach encodes 32 bits of information (so 2^32 ~= 4 billion
## possibilities) and in theory can be remapped to an integer if you
## really wanted to.
ids::sentence()

## As with `adjective_animal`, the case can be changed:
ids::sentence(2, "dot")

## If you would rather past tense for the verbs, then pass `past = TRUE`:
ids::sentence(4, past = TRUE)

## ## proquints

## "proquints" are an identifier that tries to be information dense
## but still human readable and (somewhat) pronounceable; "proquint"
## stands for *PRO*-nouncable *QUINT*-uplets.  They are introduced in
## https://arxiv.org/html/0901.4016

## `ids` can generate proquints:

ids::proquint(10)

## By default it generates two-word proquints but that can be changed:

ids::proquint(5, 1)
ids::proquint(2, 4)

## Proquints are formed by alternating
## consonant/vowel/consonant/vowel/consonant using a subset of both
## (16 consonants and 4 vowels).  This yields 2^16 (65,536)
## possibilities per word.  Words are always lower case and always
## separated by a hyphen.  So with 4 words there are 2^64 combinations
## in 23 characters.

## Proquints are also useful in that they can be tranlated with
## integers.  The proquint `kapop` has integer value 25258
ids::proquint_to_int("kapop")
ids::int_to_proquint(25258)

## This makes proquints suitable for creating human-pronouncable
## identifers out of things like ip addresses, integer primary keys,
## etc.

## The function `ids::int_to_proquint_word` will translate between
## proquint words and integers (and are vectorised)
w <- ids::int_to_proquint_word(sample(2^16, 10) - 1L)
w

## and `ids::proquint_word_to_int` does the reverse
ids::proquint_word_to_int(w)

## whille `ids::proquint_to_int` and `ids::int_to_proquint` allows
## translation of multi-word proquints.  Overflow is a real
## possibility; the maximum integer representable is only about `r
## human_no(.Machine$integer.max)` and the maximum floating point
## number of accuracy of 1 is about `r human_no(2 /
## .Machine$double.eps)` -- these are big numbers but fairly small
## proquints:
ids::int_to_proquint(.Machine$integer.max - 1)
ids::int_to_proquint(2 / .Machine$double.eps)

## But if you had a 6 word proquint this would not work!
p <- ids::proquint(1, 6)

## Too big for an integer:
##+ error = TRUE
ids::proquint_to_int(p)

## And too big for an numeric number:
##+ error = TRUE
ids::proquint_to_int(p, as = "numeric")

## To allow this, we use `openssl`'s `bignum` support:
ids::proquint_to_int(p, as = "bignum")

## This returns a *list* with one bignum (this is required to allow
## vectorisation).

## ## Roll your own identifiers

## The `ids` functions can build identifiers in the style of
## `adjective_animal` or `sentence`.  It takes as input a list of
## strings.  This works particularly well with the `rcorpora` package
## which includes lists of strings.

## Here is a list of Pokemon names:
pokemon <- tolower(rcorpora::corpora("games/pokemon")$pokemon$name)
length(pokemon)

## ...and here is a list of adjectives
adjectives <- tolower(rcorpora::corpora("words/adjs")$adjs)
length(adjectives)

## So we have a total pool size of about `r human_no(length(adjectives) *
## length(pokemon))`, which is not huge, but it is at least topical.

## To generate one identifier:
ids::ids(1, adjectives, pokemon)

## All the style-changing code is available:
ids::ids(10, adjectives, pokemon, style = "dot")

## Better would be to wrap this so that the constants are not passed
## around the whole time:
adjective_pokemon <- function(n = 1, style = "snake") {
  pokemon <- tolower(rcorpora::corpora("games/pokemon")$pokemon$name)
  adjectives <- tolower(rcorpora::corpora("words/adjs")$adjs)
  ids::ids(n, adjectives, pokemon, style = style)
}

adjective_pokemon(10, "kebab")

## As a second example we can use the word lists in rcorpora to
## generate identifiers in the form `<mood>_<scientist>`, like
## "melancholic_darwin".  These are similar to the names of
## docker containers.

## First the lists of names themselves:
moods <- tolower(rcorpora::corpora("humans/moods")$moods)
scientists <- tolower(rcorpora::corpora("humans/scientists")$scientists)

## Moods include:
sample(moods, 10)

## The scientists names contain spaces which is not going to work for
## us because `ids` won't correctly translate all internal spaces to
## the requested style.
sample(scientists, 10)

## To hack around this we'll just take the last name from the list and
## remove all hyphens:
scientists <- vapply(strsplit(sub("(-|jr\\.$)", "", scientists), " "),
                     tail, character(1), 1)

## Which gives strings that are just letters (though there are a few
## non-ASCII characters here that may cause problems because string
## handling is just a big pile of awful)
sample(scientists, 10)

## With the word lists, create an identifier:
ids::ids(1, moods, scientists)

## Or pass `NULL` for `n` and create a function:
sci_id <- ids::ids(NULL, moods, scientists, style = "kebab")

## which takes just the number of identifiers to generate as an argument
sci_id(10)

## A future version of this will keep a pool so that duplicates are
## excluded.  For now we'll just rely on the reasonably large pool.

#' Generic id generating function
#' @title Generic id generating function
#'
#' @param n number of ids to return.  If \code{NULL}, it instead
#'   returns the generating function
#'
#' @param ... A number of character vectors
#'
#' @param vals A list of character vectors, \emph{instead} of \code{...}
#'
#' @param style Style to join words with.  Can be one of "Pascal",
#'   "camel", "snake", "kebab", "dot", "title", "sentence", "lower",
#'   "upper", and "constant".
#'
#' @return Either a character vector of length \code{n}, or a
#'   function of one argument if \code{n} is \code{NULL}
#'
#' @export
#' @author Rich FitzJohn
#' @examples
#' # For an example, please see the vignette
ids <- function(n, ..., vals = list(...), style = "snake") {
  combine <- make_combine(style)
  force(vals)
  gen <- function(n = 1) {
    combine(vapply(vals, sample, character(n), n, replace = TRUE))
  }
  if (is.null(n)) gen else gen(n)
}

#' Random identifiers.  By default this uses the \code{openssl}
#' package to produce a random set of bytes, and expresses that as a
#' hex character string.  This does not affect R's random number
#' stream.
#'
#' @title Cryptographically generated random identifiers
#' @inheritParams ids
#'
#' @param bytes The number of bytes to include for each identifier.
#'   The length of the returned identifiers will be twice this long
#'   with each pair of characters representing a single byte.
#'
#' @param use_openssl A logical, indicating if we should use the
#'   openssl for generating the random identifiers.  The openssl
#'   random bytestream is not affected by the state of the R random
#'   number generator (e.g., via \code{\link{set.seed}}) so may not be
#'   suitable for use where reproducibility is important.  The speed
#'   should be very similar for both approaches.
#'
#' @export
#' @author Rich FitzJohn
#' @examples
#' # Generate a random id:
#' random_id()
#'
#' # Generate 10 of them!
#' random_id(10)
#'
#' # Different length ids
#' random_id(bytes = 8)
#' # (note that the number of characters is twice the number of bytes)
#'
#' # The ids are not affected by R's RNG state:
#' set.seed(1)
#' (id1 <- random_id())
#' set.seed(1)
#' (id2 <- random_id())
#' # The generated identifiers are different, despite the seed being the same:
#' id1 == id2
#'
#' # If you need these identifiers to be reproducible, pass use_openssl = FALSE
#' set.seed(1)
#' (id1 <- random_id(use_openssl = FALSE))
#' set.seed(1)
#' (id2 <- random_id(use_openssl = FALSE))
#' # This time they are the same:
#' id1 == id2
#'
#' # Pass \code{n = NULL} to generate a function that binds your arguments:
#' id8 <- random_id(NULL, bytes = 8)
#' id8(10)
random_id <- function(n = 1, bytes = 16, use_openssl = TRUE) {
  if (is.null(n)) {
    force(bytes)
    force(use_openssl)
    function(n = 1) {
      random_id(n, bytes, use_openssl)
    }
  } else {
    rand_bytes <- if (use_openssl) openssl::rand_bytes else r_rand_bytes
    apply(matrix(as.character(rand_bytes(n * bytes)), n), 1, paste,
          collapse = "")
  }
}

r_rand_bytes <- function(n) {
  as.raw(sample(256L, n, TRUE) - 1L)
}

context("random")

re_random <- function(len) {
  sprintf("^[[:xdigit:]]{%s}", len)
}

test_that("uuid", {
  expect_equal(random_id(0), character(0))

  x1 <- random_id(1)
  expect_equal(length(x1), 1)
  expect_is(x1, "character")
  expect_match(x1, re_random(32))

  x10 <- random_id(10)
  expect_equal(length(x10), 10)
  expect_is(x10, "character")
  expect_true(all(grepl(re_random(32), x10)))
})

test_that("bind args", {
  f <- random_id(NULL, bytes = 5)
  expect_is(f, "function")
  expect_equal(as.list(formals(f)), list(n = 1))
  expect_match(f(), re_random(10))
  expect_equal(nchar(f()), 10)
})

test_that("r byte stream", {
  set.seed(1)
  id1 <- random_id()
  set.seed(1)
  expect_false(id1 == random_id())
  set.seed(1)
  expect_false(id1 == random_id(NULL)())

  set.seed(1)
  id1 <- random_id(use_openssl = FALSE)
  set.seed(1)
  expect_true(id1 == random_id(use_openssl = FALSE))
  set.seed(1)
  expect_true(id1 == random_id(NULL, use_openssl = FALSE)())
})

## We need a bunch of constants here to keep things from being totally
## infected with magic numbers.  There are still a lot of references
## to '5' and '16' below.
PROQUINT_CONSONANT <- c("b", "d", "f", "g",
                        "h", "j", "k", "l",
                        "m", "n", "p", "r",
                        "s", "t", "v", "z")
PROQUINT_VOWEL <- c("a", "i", "o", "u")
PROQUINT_POOL <- c(PROQUINT_CONSONANT, PROQUINT_VOWEL)
PROQUINT_WORD <- 2^16 # 65536
PROQUINT_MULT <- c(16 * 4 * 16 * 4,
                   16 * 4 * 16,
                   16 * 4,
                   16,
                   1)
PROQUINT_MOD <- c(PROQUINT_WORD, PROQUINT_MULT[-5L])

PROQUINT_IDX_V <- c(2L, 4L)
PROQUINT_IDX_C <- c(1L, 3L, 5L)

PROQUINT_RE_WORD <- sprintf("[%s][%s][%s][%s][%s]",
                            paste(PROQUINT_CONSONANT, collapse = ""),
                            paste(PROQUINT_VOWEL, collapse = ""),
                            paste(PROQUINT_CONSONANT, collapse = ""),
                            paste(PROQUINT_VOWEL, collapse = ""),
                            paste(PROQUINT_CONSONANT, collapse = ""))
PROQUINT_RE <- sprintf("^%s(-%s)*$", PROQUINT_RE_WORD, PROQUINT_RE_WORD)
PROQUINT_RE1 <- sprintf("^%s$", PROQUINT_RE_WORD)

##' Generate random "proquint" identifiers.  "proquint" stands for
##' PRO-nouncable QUINT-uplets and were described by Daniel Wilkerson
##' in \url{https://arxiv.org/html/0901.4016}.  Each "word" takes one
##' of \eqn{2^16} possibilities.  A four word proquint has a keyspace
##' of \eqn{10^19} possibilities but takes only 23 characters.
##' Proquint identifiers can be interchanged with integers (though
##' this is totally optional); see \code{\link{proquint_to_int}} and
##' the other functions documented on that page.
##'
##' In the abstract of their paper, Wilkerson introduces proquints:
##'
##' "Identifiers (IDs) are pervasive throughout our modern
##' life. We suggest that these IDs would be easier to manage and
##' remember if they were easily readable, spellable, and
##' pronounceable. As a solution to this problem we propose using
##' PRO-nouncable QUINT-uplets of alternating unambiguous consonants
##' and vowels: proquints."
##'
##' @title Generate random proquint identifiers
##' @inheritParams ids
##' @param n_words The number of words for each identifier; each word
##'   has \code{2^16} (65536) possible combinations, a two-word
##'   proquint has \code{2^32} possible combinations and an
##'   \code{k}-word proquint has \code{2^(k * 16)} possible
##'   combinations.
##' @param use_cache Because there are relatively few combinations per
##'   word, and because constructing short strings is relatively
##'   expensive in R, it may be useful to cache all 65536 possible
##'   words.  If \code{TRUE} then the first time that this function is
##'   used all words will be cached and the results used - the first
##'   time may take up to ~1/4 of a second and subsequent calls will
##'   be much faster.  The identifiers selected will not change with
##'   this option (i.e., given a particular random seed, changing this
##'   option will not affect the identifiers randomly selected).
##' @param use_openssl Use openssl for random number generation, with
##'   the primary effect that the identifiers will not be affected by
##'   R's random seed (at a small speed cost).
##' @export
##' @examples
##' # A single, two word, proquint
##' proquint()
##'
##' # Longer identifier:
##' proquint(n_words = 5)
##'
##' # More identifiers
##' proquint(10)
proquint <- function(n = 1, n_words = 2L, use_cache = TRUE,
                     use_openssl = FALSE) {
  ## Consider requiring something sane for 'n_words'?
  if (is.null(n)) {
    force(n_words)
    force(use_cache)
    function(n = 1) {
      proquint(n, n_words, use_cache)
    }
  } else {
    words <- proquint_sample_words(n * n_words, use_cache, use_openssl)
    apply(matrix(words, n_words, n), 2L, paste, collapse = "-")
  }
}

##' Convert to and from proquints.
##'
##' These functions try to be type safe and predictable about what
##' they will and will not return.
##'
##' For \code{proquint_to_int}, because numeric overflow is a
##' possibility, it is important to consider whether a proquint can be
##' meaninfully translated into an integer or a numeric and the
##' functions will throw an error rather than failing in a more
##' insidious way (promoting the type or returning NA).
##'
##' \code{proquint_word_to_int} always retuns an integer vector of the
##' same length as the input.
##'
##' Missing values are allowed; a missing integer representation of a
##' proquint will translate as \code{NA_character_} and a missing
##' proquint will translate as \code{NA_integer_} (if \code{as =
##' "integer"}), \code{NA_real_}, if \code{as = "numeric"} or as
##' \code{NULL} (if \code{as = "bignum"}).
##'
##' Names are always discarded.  Future versions may gain an argument
##' \code{named} with a default of \code{FALSE}, but that setting to
##' \code{TRUE} would preseve names.  Let me know if this would be
##' useful.
##'
##' @rdname proquint_conversion
##' @title Convert to and from proquints
##' @param x An integer (or integer-like) value to convert to a
##'   proquint
##' @param p A character vector representing a proquint
##' @param i An integer representing a single proquint word (in the
##'   range 0:65535)
##' @param w A proquint \emph{word} (five letter string)
##' @inheritParams proquint
##' @export
int_to_proquint <- function(x, use_cache = TRUE) {
  if (length(x) == 0L) {
    return(character(0))
  }
  if (is.numeric(x)) {
    if (anyNA(x)) {
      return(na_recall(x, NA_character_, int_to_proquint, use_cache))
    }
    n_words <- ceiling(log(x + 1L, PROQUINT_WORD))
    if (any(n_words > 1L)) {
      scal <- PROQUINT_WORD^(rsequence(n_words) - 1L)
      idx <- rep(x, n_words) %/% scal %% PROQUINT_WORD
    } else {
      idx <- x
    }
  } else if (is.logical(x) && all(is.na(x))) {
    ## Eww nasty corner case here that turns up because a naked 'NA'
    ## in R is a logical vector.
    return(rep(NA_character_, length(x)))
  } else {
    if (inherits(x, "bignum")) {
      x <- list(x)
    } else if (!is_bignum_list(x)) {
      stop("Invalid type for 'x'")
    }
    is_na <- lengths(x) == 0
    if (any(is_na)) {
      return(na_recall(x, NA_character_, int_to_proquint, use_cache,
                       missing = is_na))
    }
    base <- openssl::bignum(PROQUINT_WORD)
    f <- function(el) {
      n_words <- big_log_ceil(el, PROQUINT_WORD)
      if (n_words == 1L) {
        as_integer_bignum(el)
      } else {
        pow <- rsequence(n_words) - 1L
        vapply(n_words - seq_len(n_words),
               function(pow) as_integer_bignum(el %/% (base^pow) %% base),
               integer(1))
      }
    }
    tmp <- lapply(x, f)
    idx <- unlist(tmp)
    n_words <- lengths(tmp)
  }

  words <- int_to_proquint_word(idx, use_cache, FALSE)

  if (any(n_words > 1L)) {
    ## It might be worth a special case here where all things are the
    ## same length because we can use matrix/apply/paste?
    i <- rep(seq_along(x), n_words)
    res <- as.character(tapply(words, i, paste, collapse = "-"))
  } else {
    res <- words
  }

  res
}

##' @param as The target data type for conversion from proquint.  The
##'   options are \code{integer}, \code{numeric} and \code{bignum}.
##'   The first two will overflow given sufficiently large input -
##'   this will throw an error (overflow is at
##'   \code{.Machine$integer.max} and \code{2 / .Machine$double.eps - 1}
##'   for \code{integer} and \code{numeric} respectively).  For
##'   \code{bignum} this will return a \emph{list} of \code{bignum}
##'   elements \emph{even if \code{p} is of length 1}.
##' @rdname proquint_conversion
##' @export
proquint_to_int <- function(p, as = "numeric", use_cache = TRUE) {
  as <- match.arg(as, c("integer", "numeric", "bignum"))
  if (length(p) == 0L) {
    return(switch(as,
                  integer = integer(0),
                  numeric = numeric(0),
                  bignum = list()))
  }
  if (anyNA(p)) {
    na <- switch(as, integer = NA_integer_, numeric = NA_real_,
                 bignum = list(NULL))
    return(na_recall(p, na, proquint_to_int, as, use_cache))
  }

  if (!is.character(p)) {
    stop("Expected a character vector for 'p'")
  }

  err <- !grepl(PROQUINT_RE, p)
  if (any(err)) {
    stop("Invalid identifier: ",
         paste(sprintf("'%s'", p[err]), collapse = ", "))
  }

  words <- strsplit(p, "-")
  len <- lengths(words)
  idx <- proquint_word_to_int(unlist(words), use_cache, FALSE)

  proquint_combine(idx, len, as)
}

##' @export
##' @rdname proquint_conversion
##' @param validate Validate the range of inputs?  Because these
##'   functions are used internally, they can skip input validation.
##'   You can too if you promise to pass sanitised input in.  If
##'   out-of-range values are passed in and validation is disabled the
##'   behaviour is undefined and subject to change.
proquint_word_to_int <- function(w, use_cache = TRUE, validate = TRUE) {
  if (anyNA(w)) {
    return(na_recall(w, NA_integer_, proquint_word_to_int,
                     use_cache, validate))
  }
  if (validate) {
    err <- !grepl(PROQUINT_RE1, w)
    if (any(err)) {
      stop(sprintf("Invalid proquint word: %s",
                   paste(sprintf("'%s'", w[err]), collapse = ", ")))
    }
  }
  if (use_cache) {
    idx <- match(w, proquint_word_cache()) - 1L
  } else {
    sx <- matrix(unlist(strsplit(unlist(w, use.names = FALSE), NULL)),
                 ncol = 5L, byrow = TRUE)
    i <- array(0L, dim(sx))
    i[, PROQUINT_IDX_C] <-
      match(sx[, PROQUINT_IDX_C], PROQUINT_CONSONANT) - 1L
    i[, PROQUINT_IDX_V] <-
      match(sx[, PROQUINT_IDX_V], PROQUINT_VOWEL) - 1L
    idx <- as.integer(drop(i %*% PROQUINT_MULT))
  }
  idx
}

##' @export
##' @rdname proquint_conversion
int_to_proquint_word <- function(i, use_cache = TRUE, validate = TRUE) {
  if (length(i) == 0) { # avoid some corner cases here
    return(character(0))
  }
  if (anyNA(i)) {
    return(na_recall(i, NA_character_, int_to_proquint_word,
                     use_cache, validate))
  }
  if (validate) {
    if (!is.numeric(i)) {
      stop("Invalid proquint word index (not numeric)")
    }
    err <- i < 0 | i >= PROQUINT_WORD
    if (any(err)) {
      stop(sprintf("Invalid proquint word index (out of range): %s",
                   paste(i[err], collapse = ", ")))
    }
  }
  if (use_cache) {
    word <- proquint_word_cache()[i + 1L]
  } else {
    j <- t(outer(i, PROQUINT_MOD, `%%`)) %/% PROQUINT_MULT
    j[PROQUINT_IDX_V, ] <- j[PROQUINT_IDX_V, ] + 16L
    word <- apply(matrix(PROQUINT_POOL[c(j) + 1L], 5L, length(i)), 2, paste0,
                  collapse = "")
  }
  word
}

## Internal support functions:
proquint_sample_words <- function(n, use_cache = TRUE, use_openssl = FALSE) {
  int_to_proquint_word(rand_i16(n, use_openssl), use_cache)
}

cache <- new.env(parent = emptyenv())
proquint_word_cache <- function() {
  if (is.null(cache$proquint_words)) {
    ## This takes ~0.25s, but would be paid across all words anyway
    idx <- as.matrix(expand.grid(1:16, 17:20, 1:16, 17:20, 1:16))[, 5:1]
    let <- array(PROQUINT_POOL[c(idx)], dim(idx))
    cache$proquint_words <- apply(let, 1, paste, collapse = "")
  }
  cache$proquint_words
}

rsequence <- function(nvec) {
  unlist(lapply(nvec, function(n) rev(seq_len(n))))
}

big_log_ceil <- function(x, base) {
  ret <- 1L
  while (x > base) {
    ret <- ret + 1L
    x <- x %/% base
  }
  ret
}

is_bignum_list <- function(x) {
  is.list(x) && all(vapply(x, function(el)
    inherits(el, "bignum") || is.null(el), logical(1)))
}

## This is a hack until openssl is updated; the development version
## has a new as.integer method that should works nicely.
##
## TODO: before CRAN release, check that this does actually work!  See
## the relevant test in test-proquint.R
as_integer_bignum <- function(x) {
  if (openssl_supports_as_integer()) {
    as.integer(x) # nocov
  } else {
    x <- as.raw(x)
    i <- length(x) - seq_along(x)
    as.integer(sum(256^i * as.integer(x)))
  }
}

openssl_supports_as_integer <- function() {
  if (is.null(cache$openssl_supports_as_integer)) {
    cache$openssl_supports_as_integer <-
      utils::packageVersion("openssl") > "0.9.6"
  }
  cache$openssl_supports_as_integer
}

na_recall <- function(x, na, fun, ..., missing = is.na(x)) {
  ret <- rep(na, length(x))
  i <- !missing
  ret[i] <- fun(x[i], ...)
  ret
}

## This tries to deal with the nastiness of overflowing but keeps most
## of the sausage factory out of the main functions.
proquint_combine <- function(idx, len, as) {
  grp <- rep(seq_along(len), len)

  if (as == "bignum") {
    big_combine1 <- function(x) {
      n <- length(x)
      base <- openssl::bignum(PROQUINT_WORD)
      res <- openssl::bignum(0)
      for (i in seq_along(x)) {
        res <- res + x[[i]] * base^(n - i)
      }
      res
    }
    res <- tapply(idx, grp, big_combine1)
    attributes(res) <- NULL
  } else {
    scal <- PROQUINT_WORD^(rsequence(len) - 1L)
    res <- tapply(scal * idx, rep(seq_along(len), len), sum)
    if (as == "integer") {
      i <- res > .Machine$integer.max
      if (any(i)) {
        stop("Integer overflow: cannot represent proquint as integer")
      }
      res <- as.integer(res)
    } else {
      i <- res >= 2 / .Machine$double.eps
      if (any(i)) {
        stop("Numeric overflow: cannot represent proquint as numeric")
      }
      res <- as.numeric(res)
    }
  }
  res
}

rand_i16 <- function(n, use_openssl = FALSE) {
  if (use_openssl) {
    r <- matrix(as.integer(openssl::rand_bytes(2 * n)), 2)
    r[1L, ] <- r[1L, ] * 256L
    as.integer(colSums(matrix(r, 2)))
  } else {
    sample(PROQUINT_WORD, n, replace = TRUE) - 1L
  }
}

context("proquint")

test_that("word conversions", {
  i <- 10^(0:4)
  ## From Python:
  cmp <- c("babad", "babap", "badoh", "bazom", "fisib")
  ## Ours
  w <- int_to_proquint_word(i)
  expect_identical(w, cmp)

  j <- proquint_word_to_int(w, TRUE)
  expect_equal(j, i)
  expect_is(j, "integer")

  k <- proquint_word_to_int(w, FALSE)
  expect_identical(j, k)
  expect_is(k, "integer")
})

## This is exhaustive, because it's not that slow to do everything:
test_that("word conversions - exhaustive", {
  i <- seq_len(PROQUINT_WORD) - 1L
  w <- int_to_proquint_word(i)
  expect_equal(w, cache$proquint_words)
  expect_equal(int_to_proquint_word(i, FALSE), cache$proquint_words)
  ## word -> index
  expect_equal(proquint_word_to_int(w), i)
  expect_equal(proquint_word_to_int(w, FALSE), i)
})

test_that("proquint conversions", {
  ## Then start on the multi-word cases
  known <- c("babad" = 1,
             "babap" = 10,
             "badoh" = 100,
             "bazom" = 1000,
             "fisib" = 10000,
             "babad-mipob" = 100000,
             "babaz-hanab" = 1000000,
             "bafim-nipab" = 10000000,
             "biluj-vahab" = 100000000,
             "govip-somab" = 1000000000)

  expect_equal(proquint_to_int(names(known), use_cache = TRUE),
               unname(known))
  expect_equal(proquint_to_int(names(known), use_cache = FALSE),
               unname(known))

  ## Then the reverse:
  expect_equal(int_to_proquint(unname(known), TRUE), names(known))
  expect_equal(int_to_proquint(unname(known), FALSE), names(known))
})

## Around the transition:
test_that("1-2 word transition corner cases", {
  i <- (PROQUINT_WORD - 2):(PROQUINT_WORD + 2)
  expected <- c("zuzuv", "zuzuz", "babad-babab", "babad-babad", "babad-babaf")
  expect_equal(int_to_proquint(i), expected)
  expect_equal(int_to_proquint(i, FALSE), expected)
  expect_equal(proquint_to_int(expected, use_cache = TRUE), i)
  expect_equal(proquint_to_int(expected, use_cache = FALSE), i)
})

test_that("sampled words do not depend on cache", {
  set.seed(1)
  w1 <- proquint_sample_words(10, TRUE)
  set.seed(1)
  w2 <- proquint_sample_words(10, FALSE)
  expect_identical(w1, w2)
})

test_that("generate openssl random numbers", {
  i <- rand_i16(1, TRUE)
  expect_is(i, "integer")
  expect_gte(i, 0L)
  expect_lt(i, 2^16)

  j <- rand_i16(10, TRUE)
  expect_is(j, "integer")
  expect_equal(length(j), 10)
  expect_true(all(j >= 0L))
  expect_true(all(j < 2^16))

  set.seed(1)
  i <- rand_i16(100, TRUE)
  set.seed(1)
  j <- rand_i16(100, TRUE)
  expect_false(identical(i, j))

  i <- rand_i16(2^16 * 16, TRUE)
  n <- tabulate(i + 1, 2^16)
  expect_true(all(i >= 0L))
  expect_true(all(i < 2^16))
})

test_that("openssl random identifiers", {
  set.seed(1)
  w1 <- proquint(100, use_openssl = TRUE)
  set.seed(1)
  w2 <- proquint(100, use_openssl = TRUE)
  expect_false(identical(w1, w2))
})

test_that("bad type on conversion", {
  expect_error(int_to_proquint("one"), "Invalid type for 'x'")
  expect_error(int_to_proquint(TRUE), "Invalid type for 'x'")
  expect_error(int_to_proquint(1+4i), "Invalid type for 'x'")
})

test_that("word -> int translation: missing values", {
  i <- 12345L
  w <- int_to_proquint_word(i)
  expect_identical(proquint_word_to_int(NA), NA_integer_)
  expect_identical(proquint_word_to_int(NA_character_), NA_integer_)
  expect_identical(proquint_word_to_int(c(NA_character_, w)),
                   c(NA_integer_, i))
  expect_identical(proquint_word_to_int(c(NA, w, w, NA)),
                   c(NA_integer_, i, i, NA_integer_))
})

test_that("int -> word translation: missing values", {
  i <- 12345L
  w <- int_to_proquint_word(i)
  expect_identical(int_to_proquint_word(NA), NA_character_)
  expect_identical(int_to_proquint_word(NA_integer_), NA_character_)
  expect_identical(int_to_proquint_word(c(NA_integer_, i)),
                   c(NA_character_, w))
  expect_identical(int_to_proquint_word(c(NA, i, i, NA)),
                   c(NA_character_, w, w, NA_character_))
})

test_that("identifier -> int translation: missing values", {
  i <- 12345L
  w <- int_to_proquint_word(i)
  ib <- openssl::bignum(i)

  expect_identical(proquint_to_int(NA, "integer"), NA_integer_)
  expect_identical(proquint_to_int(NA, "numeric"), NA_real_)
  expect_identical(proquint_to_int(NA, "bignum"), list(NULL))

  expect_identical(proquint_to_int(NA_character_, "integer"), NA_integer_)
  expect_identical(proquint_to_int(NA_character_, "numeric"), NA_real_)
  expect_identical(proquint_to_int(NA_character_, "bignum"), list(NULL))

  expect_identical(proquint_to_int(c(NA, w), "integer"), c(NA_integer_, i))
  expect_identical(proquint_to_int(c(NA, w), "numeric"), c(NA_real_, i))
  expect_identical(proquint_to_int(c(NA, w), "bignum"),
                   list(NULL, ib))

  expect_identical(proquint_to_int(c(NA, w, w, NA), "integer"),
                   c(NA_integer_, i, i, NA_integer_))
  expect_identical(proquint_to_int(c(NA, w, w, NA), "numeric"),
                   c(NA_real_, i, i, NA_real_))
  expect_identical(proquint_to_int(c(NA, w, w, NA), "bignum"),
                   list(NULL, ib, ib, NULL))
})

test_that("int -> identifier translation: missing values", {
  i <- 12345L
  w <- int_to_proquint_word(i)
  ib <- openssl::bignum(i)

  expect_identical(int_to_proquint(NA), NA_character_)
  expect_identical(int_to_proquint(NA_integer_), NA_character_)
  expect_identical(int_to_proquint(NA_real_), NA_character_)
  expect_identical(int_to_proquint(list(NULL)), NA_character_)

  expect_identical(int_to_proquint(c(NA_integer_, i)), c(NA, w))
  expect_identical(int_to_proquint(c(NA_real_, i)), c(NA, w))
  expect_identical(int_to_proquint(list(NULL, ib)), c(NA, w))

  expect_identical(int_to_proquint(c(NA_integer_, i, i, NA_integer_)),
                   c(NA_character_, w, w, NA_character_))
  expect_identical(int_to_proquint(c(NA_real_, i, i, NA_real_)),
                   c(NA_character_, w, w, NA_character_))
  expect_identical(int_to_proquint(list(NULL, ib, ib, NULL)),
                   c(NA_character_, w, w, NA_character_))
})

test_that("identifier: 0", {
  expect_identical(proquint(0), character(0))
})

test_that("identifier: 1", {
  x <- proquint(1)
  expect_is(x, "character")
  expect_equal(length(x), 1)
  re <- sprintf("^%s-%s$", PROQUINT_RE_WORD, PROQUINT_RE_WORD)
  expect_match(x, re)
})

test_that("identifier: n", {
  n <- 10
  x <- proquint(n)
  expect_is(x, "character")
  expect_equal(length(x), n)
  re <- sprintf("^%s-%s$", PROQUINT_RE_WORD, PROQUINT_RE_WORD)
  expect_match(x, re, all = TRUE)
})

test_that("vary word length", {
  re3 <- sprintf("^%s-%s-%s$",
                 PROQUINT_RE_WORD, PROQUINT_RE_WORD, PROQUINT_RE_WORD)
  expect_match(proquint(1, 3), re3)
  expect_match(proquint(1, 1), PROQUINT_RE1)
})

test_that("functional interface", {
  re3 <- sprintf("^%s-%s-%s$",
                 PROQUINT_RE_WORD, PROQUINT_RE_WORD, PROQUINT_RE_WORD)
  f <- proquint(NULL, 3)
  expect_is(f, "function")
  expect_match(f(), re3)
  expect_equal(length(f()), 1)
  expect_equal(length(f(10)), 10)
})

## -- below here is all boring error handling stuff

## Validate words
test_that("invalid word -> int", {
  expect_error(proquint_word_to_int("hello!"),
               "Invalid proquint word: 'hello!'")
  expect_error(proquint_word_to_int(c("hello!", "babad"),),
               "Invalid proquint word: 'hello!'")
  expect_error(proquint_word_to_int(c("hello!", "babad", "yo!"),),
               "Invalid proquint word: 'hello!', 'yo!'")
  expect_error(proquint_word_to_int(1),
               "Invalid proquint word: '1'")
})

## Validate word indexes
test_that("invalid word -> int", {
  expect_error(int_to_proquint_word(-1),
               "Invalid proquint word index (out of range): -1",
               fixed = TRUE)
  expect_error(int_to_proquint_word(c(-1, 1)),
               "Invalid proquint word index (out of range): -1",
               fixed = TRUE)
  expect_error(int_to_proquint_word(c(-1, 1, 65536)),
               "Invalid proquint word index (out of range): -1, 65536",
               fixed = TRUE)
  expect_error(int_to_proquint_word("babad"),
               "Invalid proquint word index (not numeric)",
               fixed = TRUE)
})

test_that("zero length input", {
  expect_equal(proquint_word_to_int(NULL), integer(0))
  expect_equal(proquint_word_to_int(character(0)), integer(0))
  ## This is not terrific but it is what it is:
  expect_equal(proquint_word_to_int(integer(0)), integer(0))

  expect_equal(int_to_proquint_word(NULL), character(0))
  expect_equal(int_to_proquint_word(integer(0)), character(0))
})

test_that("zero length input", {
  expect_identical(proquint_to_int(NULL, "integer"), integer(0))
  expect_identical(proquint_to_int(NULL, "numeric"), numeric(0))
  expect_identical(proquint_to_int(NULL, "bignum"), list())

  expect_identical(proquint_to_int(character(0), "integer"), integer(0))
  expect_identical(proquint_to_int(character(0), "numeric"), numeric(0))
  expect_identical(proquint_to_int(character(0), "bignum"), list())
})

test_that("zero length input", {
  expect_identical(int_to_proquint(NULL), character(0))
  expect_identical(int_to_proquint(logical(0)), character(0))
  expect_identical(int_to_proquint(integer(0)), character(0))
  expect_identical(int_to_proquint(numeric(0)), character(0))
  expect_identical(int_to_proquint(list()), character(0))
})

test_that("invalid identifier", {
  expect_error(proquint_to_int(TRUE), "Expected a character vector for 'p'")
  expect_error(proquint_to_int("aaaaa"), "Invalid identifier: 'aaaaa'")
})

test_that("proquint to int, varying formats", {
  i <- c(1, 10, 100, 1000, 10000)
  w <- int_to_proquint_word(i)
  expect_identical(proquint_to_int(w, "integer"), as.integer(i))
  expect_identical(proquint_to_int(w, "numeric"), as.numeric(i))
  expect_identical(proquint_to_int(w, "bignum"),
                   lapply(i, openssl::bignum))
})

test_that("int to proquint, varying formats", {
  i <- c(1, 10, 100, 1000, 10000)
  w <- int_to_proquint_word(i)
  expect_identical(int_to_proquint(as.integer(i)), w)
  expect_identical(int_to_proquint(as.numeric(i)), w)
  expect_identical(int_to_proquint(lapply(i, openssl::bignum)), w)
})

test_that("integer overflow", {
  big <- .Machine$integer.max * 2
  pq <- int_to_proquint(big)
  expect_is(pq, "character")
  expect_error(proquint_to_int(pq, "integer"),
               "Integer overflow: cannot represent proquint as integer")
  expect_identical(proquint_to_int(pq, "numeric"), big)
  expect_identical(proquint_to_int(pq, "bignum"),
                   list(openssl::bignum(big)))
})

test_that("numeric overflow", {
  pow <- log2(2/.Machine$double.eps) + 1
  big <- openssl::bignum(2)^pow + 1

  pq <- int_to_proquint(big)
  expect_is(pq, "character")
  expect_error(proquint_to_int(pq, "integer"),
               "Integer overflow: cannot represent proquint as integer")
  expect_error(proquint_to_int(pq, "numeric"),
               "Numeric overflow: cannot represent proquint as numeric")
  expect_identical(proquint_to_int(pq, "bignum"),
                   list(openssl::bignum(big)))
})

## This supports my temporary as_integer_bignum function:
test_that("as_integer_bignum", {
  x <- c(0L, 255L, 256L, 1000L, 200000L, .Machine$integer.max)
  for (i in x) {
    expect_identical(as_integer_bignum(openssl::bignum(i)), i)
  }
})

# PascalCase
# camelCase
# snake_case
# kebab-case
# Title Case
# lower case
# upper case
# CONSTANT_CASE
# Sentence case
check_style <- function(style) {
  if (!is.character(style)) {
    stop("style must be a character vector")
  }
  if (length(style) != 1L) {
    stop("style must be a scalar")
  }
  pos <- cases()
  i <- pmatch(sub("[_-]?case$", "", tolower(style)), names(pos))
  if (is.na(i)) {
    stop(sprintf("Invalid style '%s'", style))
  }
  ret <- pos[[i]]
  ret$name <- names(pos)[[i]]
  ret
}

make_combine <- function(style) {
  dat <- check_style(style)
  join <- dat$join
  tr <- dat$tr
  function(x) {
    if (is.matrix(x)) {
      apply(tr(x), 1, paste, collapse = join)
    } else {
      paste(tr(x), collapse = join)
    }
  }
}

toupper_initial <- function(x) {
  substr(x, 1, 1) <- toupper(substr(x, 1, 1))
  x
}

toupper_camel <- function(x) {
  if (is.matrix(x)) {
    x[, -1] <- toupper_initial(x[, -1])
  } else {
    x[-1] <- toupper_initial(x[-1])
  }
  x
}

toupper_sentence <- function(x) {
  if (is.matrix(x)) {
    x[, 1] <- toupper_initial(x[, 1])
  } else {
    x[1] <- toupper_initial(x[1])
  }
  x
}

toupper_pascal <- function(x) {
  x[] <- toupper_initial(x)
  x
}

cases <- function() {
  list(snake    = list(join = "_", tr = identity),
       kebab    = list(join = "-", tr = identity),
       dot      = list(join = ".", tr = identity),
       camel    = list(join = "",  tr = toupper_camel),
       pascal   = list(join = "",  tr = toupper_pascal),
       lower    = list(join = " ", tr = identity), # lower case already
       upper    = list(join = " ", tr = toupper),
       title    = list(join = " ", tr = toupper_pascal),
       sentence = list(join = " ", tr = toupper_sentence),
       constant = list(join = "_", tr = toupper))
}

context("case")

test_that("style match", {
  expect_equal(check_style("p")$name, "pascal")
  expect_equal(check_style("pascal")$name, "pascal")
  expect_equal(check_style("pascalcase")$name, "pascal")
  expect_equal(check_style("PascalCase")$name, "pascal")
  expect_equal(check_style("Pascal-Case")$name, "pascal")
  expect_equal(check_style("Pascal_Case")$name, "pascal")
})

test_that("camel", {
  expect_equal(toupper_camel(character(0)), character(0))
  expect_equal(toupper_camel("aaa"), "aaa")
  expect_equal(toupper_camel(c("aaa", "bbb")), c("aaa", "Bbb"))
  expect_equal(toupper_camel(rbind(c("aaa", "bbb"),
                                   c("ccc", "ddd"))),
               rbind(c("aaa", "Bbb"),
                     c("ccc", "Ddd")))
})

test_that("sentence", {
  tr <- make_combine("sentence")
  expect_equal(tr(c("foo", "bar")), "Foo bar")
  expect_equal(tr(rbind(c("foo", "bar"),
                        c("another", "sentence"))),
               c("Foo bar",
                 "Another sentence"))
})

test_that("error cases", {
  expect_error(check_style(NULL), "must be a character vector")
  expect_error(check_style(1), "must be a character vector")
  expect_error(check_style(character(0)), "must be a scalar")
  expect_error(check_style(c("camel", "upper")), "must be a scalar")
  expect_error(check_style("unknown"), "Invalid style 'unknown'")
})

context("ids")

test_that("animals", {
  re_snake2 <- "^([a-z]+)_([a-z]+)$"
  re_camel2 <- "^([a-z]+)([A-Z][a-z]+)$"
  re_snake3 <- "^([a-z]+)_([a-z]+)_([a-z]+)$"

  res <- adjective_animal()
  expect_match(res, re_snake2)
  expect_match(adjective_animal(1, 2), re_snake3)

  expect_match(adjective_animal(style = "camel"), re_camel2)
  res2 <- adjective_animal(2, style = "camel")
  expect_match(res2, re_camel2)
  expect_true(all(sub(re_camel2, "\\1", res2) %in% gfycat_adjectives))
  expect_true(all(tolower(sub(re_camel2, "\\2", res2)) %in% gfycat_animals))

  ## Smoke test:
  for (s in names(cases())) {
    expect_is(adjective_animal(style = s), "character")
  }
})

#' Ids based on a number of adjectives and an animal
#'
#' The list of adjectives and animals comes from
#' \url{https://github.com/a-type/adjective-adjective-animal}, and in
#' turn from \url{gfycat.com}
#'
#' @title Ids based on a number of adjectives and an animal
#' @param n_adjectives Number of adjectives to prefix the anmial with
#'
#' @param max_len The maximum length of a word part to include (this
#'   may be useful because some of the names are rather long.  This
#'   stops you generating a
#'   \code{hexakosioihexekontahexaphobic_queenalexandrasbirdwingbutterfly}).
#'   A vector of length 2 can be passed in here in which case the
#'   first element will apply to the adjectives (all of them) and the
#'   second element will apply to the animals.
#'
#' @inheritParams ids
#' @export
#' @author Rich FitzJohn
#' @examples
#' # Generate a random identifier:
#' adjective_animal()
#'
#' # Generate a bunch all at once:
#' adjective_animal(5)
#'
#' # Control the style of punctuation with the style argument:
#' adjective_animal(style = "lower")
#' adjective_animal(style = "CONSTANT")
#' adjective_animal(style = "camel")
#' adjective_animal(style = "kebab")
#'
#' # Control the number of adjectives used
#' adjective_animal(n_adjectives = 3)
#'
#' # This can get out of hand quickly though:
#' adjective_animal(n_adjectives = 7)
#'
#' # Limit the length of adjectives and animals used:
#' adjective_animal(10, max_len = 6)
#'
#' # The lengths can be controlled for adjectives and animals
#' # separately, with Inf meaning no limit:
#' adjective_animal(10, max_len = c(6, Inf), n_adjectives = 2)
#'
#' # Pass n = NULL to bind arguments to a function
#' id <- adjective_animal(NULL, n_adjectives = 2, style = "dot", max_len = 6)
#' id()
#' id(10)
adjective_animal <- function(n = 1, n_adjectives = 1, style = "snake",
                             max_len = Inf) {
  if (any(is.finite(max_len))) {
    if (any(max_len < 3)) {
      stop("max_len must be at least 3")
    }
    if (length(max_len) == 1L) {
      max_len <- rep_len(max_len, 2)
    } else if (length(max_len) != 2L) {
      stop("max_len must be length 1 or 2")
    }
    gfycat_adjectives <-
      gfycat_adjectives[nchar(gfycat_adjectives) <= max_len[1]]
    gfycat_animals <- gfycat_animals[nchar(gfycat_animals) <= max_len[2]]
  }
  vals <- c(rep(list(gfycat_adjectives), n_adjectives), list(gfycat_animals))
  ids(n, vals = vals, style = style)
}

context("adjective animal")

test_that("data", {
  expect_false(any(grepl("\\s", gfycat_animals)))
  expect_false(any(grepl("\\s", gfycat_adjectives)))
})

test_that("basic", {
  res <- adjective_animal()
  expect_is(res, "character")
  expect_equal(length(res), 1)
  expect_match(res, "^[a-z]+_[a-z]+$")
})

test_that("length", {
  expect_equal(length(adjective_animal(5)), 5)
  expect_equal(adjective_animal(0), character(0))
})

test_that("style", {
  expect_match(adjective_animal(style = "Pascal"), "^[A-Z][a-z]+[A-Z][a-z]+$")
  expect_match(adjective_animal(style = "camel"), "^[a-z]+[A-Z][a-z]+$")
  expect_match(adjective_animal(style = "snake"), "^[a-z]+_[a-z]+$")
  expect_match(adjective_animal(style = "kebab"), "^[a-z]+-[a-z]+$")
  expect_match(adjective_animal(style = "lower"), "^[a-z]+ [a-z]+$")
  expect_match(adjective_animal(style = "upper"), "^[A-Z]+ [A-Z]+$")
  expect_match(adjective_animal(style = "constant"), "^[A-Z]+_[A-Z]+$")
  expect_match(adjective_animal(style = "title"), "^[A-Z][a-z]+ [A-Z][a-z]+$")
  expect_match(adjective_animal(style = "sentence"), "^[A-Z][a-z]+ [a-z]+$")
})

test_that("restrict length", {
  tmp <- strsplit(adjective_animal(200, max_len = 5), "_", fixed = TRUE)
  expect_lte(max(sapply(tmp, nchar)), 5)

  tmp <- strsplit(adjective_animal(200, max_len = c(5, 10)), "_", fixed = TRUE)
  tmp <- apply(sapply(tmp, nchar), 1, max)
  expect_lte(tmp[[1]], 5)
  expect_lte(tmp[[2]], 10)
  expect_gt(tmp[[2]], tmp[[1]])
})

test_that("restrict length error cases", {
  expect_error(adjective_animal(max_len = 2),
               "max_len must be at least 3")
  expect_error(adjective_animal(max_len = -2),
               "max_len must be at least 3")

  ## This is ignored -- perhaps it should not be?
  expect_silent(adjective_animal(max_len = numeric(0)))

  expect_error(adjective_animal(max_len = c(5, 6, 7)),
               "max_len must be length 1 or 2")
  expect_error(adjective_animal(n_adjectives = 2, max_len = c(5, 6, 7)),
               "max_len must be length 1 or 2")
})

test_that("functional interface", {
  f <- adjective_animal(NULL, max_len = 10, style = "kebab", n_adjectives = 2)
  expect_is(f, "function")
  re <- "^[a-z]{2,10}-[a-z]{2,10}-[a-z]{2,10}$"
  expect_match(f(), re)
  x <- f(100)
  expect_true(all(grepl(re, x)))
})

#' Create a sentence style identifier.  This uses the approach
#' described by Asana on their blog
#' \url{https://blog.asana.com/2011/09/6-sad-squid-snuggle-softly/}.
#' This approach encodes 32 bits of information (so 2^32 ~= 4 billion
#' possibilities) and in theory can be remapped to an integer if you
#' really wanted to.
#'
#' @title Sentence style identifiers
#'
#' @inheritParams ids
#'
#' @param past Use the past tense for verbs (e.g., slurped or jogged
#'   rather than slurping or jogging)
#'
#' @export
#' @author Rich FitzJohn
#' @examples
#' # Generate an identifier
#' sentence()
#'
#' # Generate a bunch
#' sentence(10)
#'
#' # As with adjective_animal, use "style" to control punctuation
#' sentence(style = "Camel")
#' sentence(style = "dot")
#' sentence(style = "Title")
#'
#' # Change the tense of the verb:
#' set.seed(1)
#' sentence()
#' set.seed(1)
#' sentence(past = TRUE)
#'
#' # Pass n = NULL to bind arguments to a function
#' id <- sentence(NULL, past = TRUE, style = "dot")
#' id()
#' id(10)
sentence <- function(n = 1, style = "snake", past = FALSE) {
  verbs <- if (past) asana_verbs_past else asana_verbs_present
  ids(n, asana_ids, asana_adjectives, asana_nouns, verbs, asana_adverbs,
      style = style)
}

asana_ids <- as.character(2:33)

asana_adjectives <- c(
  "adorable", "adventurous", "alluring", "amazing",
  "ambitious", "amusing", "astonishing", "attractive", "awesome",
  "bashful", "bawdy", "beautiful", "bewildered", "bizarre", "bouncy",
  "brainy", "brave", "brawny", "burly", "capricious", "careful",
  "caring", "cautious", "charming", "cheerful", "chivalrous",
  "classy", "clever", "clumsy", "colossal", "cool", "coordinated",
  "courageous", "cuddly", "curious", "cute", "daffy", "dapper",
  "dashing", "dazzling", "delicate", "delightful", "determined",
  "eager", "embarrassed", "enchanted", "energetic", "enormous",
  "entertaining", "enthralling", "enthusiastic", "evanescent",
  "excited", "exotic", "exuberant", "exultant", "fabulous", "fancy",
  "festive", "finicky", "flashy", "flippant", "fluffy", "fluttering",
  "funny", "furry", "fuzzy", "gaudy", "gentle", "giddy", "glamorous",
  "gleaming", "goofy", "gorgeous", "graceful", "grandiose", "groovy",
  "handsome", "happy", "hilarious", "honorable", "hulking",
  "humorous", "industrious", "incredible", "intelligent", "jazzy",
  "jolly", "joyous", "kind", "macho", "magnificent", "majestic",
  "marvelous", "mighty", "mysterious", "naughty", "nimble", "nutty",
  "oafish", "obnoxious", "outrageous", "pretty", "psychedelic",
  "psychotic", "puzzled", "quirky", "quizzical", "rambunctious",
  "remarkable", "sassy", "shaggy", "smelly", "sneaky", "spiffy",
  "swanky", "sweet", "swift", "talented", "thundering", "unkempt",
  "upbeat", "uppity", "wacky", "waggish", "whimsical", "wiggly",
  "zany")

asana_nouns <- c(
  "aardvarks", "alligators", "alpacas", "anteaters", "antelopes",
  "armadillos", "baboons", "badgers", "bears", "beavers",
  "boars", "buffalos", "bulls", "bunnies", "camels", "cats",
  "chameleons", "cheetahs", "centaurs", "chickens", "chimpanzees",
  "chinchillas", "chipmunks", "cougars", "cows", "coyotes", "cranes",
  "crickets", "crocodiles", "deers", "dinasaurs", "dingos", "dogs",
  "donkeys", "dragons", "elephants", "elves", "ferrets", "flamingos",
  "foxes", "frogs", "gazelles", "giraffes", "gnomes", "gnus", "goats",
  "gophers", "gorillas", "hamsters", "hedgehogs", "hippopotamus",
  "hobbits", "hogs", "horses", "hyenas", "ibexes", "iguanas",
  "impalas", "jackals", "jackalopes", "jaguars", "kangaroos",
  "kittens", "koalas", "lambs", "lemmings", "leopards", "lions",
  "ligers", "lizards", "llamas", "lynxes", "meerkat", "moles",
  "mongooses", "monkeys", "moose", "mules", "newts", "okapis",
  "orangutans", "ostriches", "otters", "oxes", "pandas", "panthers",
  "peacocks", "pegasuses", "phoenixes", "pigeons", "pigs",
  "platypuses", "ponies", "porcupines", "porpoises", "pumas",
  "pythons", "rabbits", "raccoons", "rams", "reindeers",
  "rhinoceroses", "salamanders", "seals", "sheep", "skunks",
  "sloths", "slugs", "snails", "snakes", "sphinxes", "sprites",
  "squirrels", "takins", "tigers", "toads", "trolls", "turtles",
  "unicorns", "walruses", "warthogs", "weasels", "wolves",
  "wolverines", "wombats", "woodchucks", "yaks", "zebras")

asana_verbs_past <- c(
  "ambled", "assembled", "burst", "babbled", "charged", "chewed",
  "clamored", "coasted", "crawled", "crept", "danced", "dashed",
  "drove", "flopped", "galloped", "gathered", "glided", "hobbled",
  "hopped", "hurried", "hustled", "jogged", "juggled", "jumped",
  "laughed", "marched", "meandered", "munched", "passed", "plodded",
  "pranced", "ran", "raced", "rushed", "sailed", "sang", "sauntered",
  "scampered", "scurried", "skipped", "slogged", "slurped", "spied",
  "sprinted", "spurted", "squiggled", "squirmed", "stretched",
  "strode", "strut", "swam", "swung", "traveled", "trudged",
  "tumbled", "twisted", "wade", "wandered", "whistled", "wiggled",
  "wobbled", "yawned", "zipped", "zoomed")

asana_verbs_present <- c(
  "ambling", "assembling", "bursting", "babbling", "charging", "chewing",
  "clamoring", "coasting", "crawling", "creeping", "dancing", "dashing",
  "driving", "flopping", "galloping", "gathering", "gliding", "hobbling",
  "hopping", "hurrying", "hustling", "jogging", "juggling", "jumping",
  "laughing", "marching", "meandering", "munching", "passing", "plodding",
  "prancing", "running", "racing", "rushing", "sailing", "singing",
  "sauntering", "scampering", "scurriing", "skipping", "slogging",
  "slurping", "spying", "sprinting", "spurting", "squiggling", "squirming",
  "stretching", "striding", "strutting", "swimming", "swinging", "traveling",
  "trudging", "tumbling", "twisting", "wading", "wandering", "whistling",
  "wiggling", "wobbling", "yawning", "zipping", "zooming")

asana_adverbs <- c(
  "absentmindedly", "adventurously", "angrily", "anxiously",
  "awkwardly", "bashfully", "beautifully", "bleakly", "blissfully",
  "boastfully", "boldly", "bravely", "briskly", "calmly",
  "carefully", "cautiously", "cheerfully", "cleverly", "cluelessly",
  "clumsily", "coaxingly", "colorfully", "coolly", "courageously",
  "curiously", "daintily", "defiantly", "deliberately",
  "delightfully", "diligently", "dreamily", "drudgingly", "eagerly",
  "effortlessly", "elegantly", "energetically", "enthusiastically",
  "excitedly", "fervently", "foolishly", "furiously", "gallantly",
  "gently", "gladly", "gleefully", "gracefully", "gratefully",
  "happily", "hastily", "haphazardly", "hungrily", "innocently",
  "inquisitively", "intensely", "jokingly", "joshingly", "joyously",
  "jovially", "jubilantly", "kiddingly", "knavishly", "knottily",
  "kookily", "lazily", "loftily", "longingly", "lovingly", "loudly",
  "loyally", "madly", "majestically", "merrily", "mockingly",
  "mysteriously", "nervously", "noisily", "obnoxiously", "oddly",
  "optimistically", "overconfidently", "outside", "owlishly",
  "patiently", "playfully", "politely", "powerfully", "purposefully",
  "quaintly", "quarrelsomely", "queasily", "quickly", "quietly",
  "quirkily", "quizzically", "rapidly", "reassuringly", "recklessly",
  "reluctantly", "reproachfully", "sadly", "scarily", "seriously",
  "shakily", "sheepishly", "shyly", "silently", "sillily",
  "sleepily", "slowly", "speedily", "stealthily", "sternly",
  "suspiciously", "sweetly", "tenderly", "tensely", "thoughtfully",
  "triumphantly", "unabashedly", "unaccountably", "urgently",
  "vainly", "valiantly", "victoriously", "warmly", "wearily",
  "youthfully", "zestfully")

context("sentence")

test_that("basic", {
  re <- "^[0-9]+(_[a-z]+){4}$"
  res <- sentence()
  expect_is(res, "character")
  expect_equal(length(res), 1)
  expect_match(res, re)
})

test_that("tense", {
  verb <- vapply(strsplit(sentence(100), "_", fixed = TRUE),
                 function(x) x[[4L]], character(1))
  expect_true(all(verb %in% asana_verbs_present))

  verb <- vapply(strsplit(sentence(100, past = TRUE), "_", fixed = TRUE),
                 function(x) x[[4L]], character(1))
  expect_true(all(verb %in% asana_verbs_past))
})

test_that("functional interface", {
  f <- sentence(NULL, style = "kebab", past = TRUE)
  expect_is(f, "function")
  re <- "^[0-9]+(-[a-z]+){4}$"
  expect_match(f(), re)
  res <- f(100)
  expect_true(all(grepl(re, res)))

  verb <- vapply(strsplit(res, "-", fixed = TRUE),
                 function(x) x[[4L]], character(1))
  expect_true(all(verb %in% asana_verbs_past))
})

#' Generate UUIDs using the uuid package.  This is simply a thin
#' wrapper around \code{uuid::UUIDgenerate} that matches the interface
#' in the rest of the ids package.
#'
#' @title Generate UUIDs
#'
#' @inheritParams ids
#'
#' @param drop_hyphens Drop the hyphens from the UUID?
#'
#' @param use_time Passed through to \code{UUIDgenerate} as \code{use.time}.
#'
#' @export
#'
#' @importFrom uuid UUIDgenerate
#' @author Rich FitzJohn
#' @examples
#' # Generate one id
#' uuid()
#'
#' # Or a bunch
#' uuid(10)
#'
#' # More in the style of random_id()
#' uuid(drop_hyphens = TRUE)
uuid <- function(n = 1, drop_hyphens = FALSE, use_time = NA) {
  if (is.null(n)) {
    force(drop_hyphens)
    force(use_time)
    function(n = 1) {
      uuid(n, drop_hyphens, use_time)
    }
  } else {
    res <- vapply(seq_len(n), function(i) uuid::UUIDgenerate(use_time),
                  character(1))
    if (drop_hyphens) gsub("-", "", res, fixed = TRUE) else res
  }
}

context("uuid")

RE_UUID <- "^[[:xdigit:]]{8}-([[:xdigit:]]{4}-){3}[[:xdigit:]]{12}$"

test_that("uuid", {
  expect_equal(uuid(0), character(0))

  x1 <- uuid(1)
  expect_equal(length(x1), 1)
  expect_is(x1, "character")
  expect_match(x1, RE_UUID)

  x10 <- uuid(10)
  expect_equal(length(x10), 10)
  expect_is(x10, "character")
  expect_true(all(grepl(RE_UUID, x10)))

  expect_true(all(grepl("^[[:xdigit:]]{32}$", uuid(10, drop_hyphens = TRUE))))

  expect_match(uuid(use_time = TRUE), RE_UUID)
})

test_that("bind args", {
  f <- uuid(NULL)
  expect_is(f, "function")
  expect_equal(as.list(formals(f)), list(n = 1))
  expect_match(f(), RE_UUID)

  g <- uuid(NULL, TRUE)
  expect_match(g(), "^[[:xdigit:]]{32}$")
})

---
title: "ids"
author: "Rich FitzJohn"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{ids}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

``` {r echo = FALSE, results = "hide"}
knitr::opts_chunk$set(error = FALSE)
human_no <- function(x) {
  s <- log10(floor(x + 1))
  p <- c(0, thousand = 3, million = 6, billion = 9, trillion = 12)
  i <- s > p
  j <- max(which(i))
  str <- names(p)[j]
  if (nzchar(str)) {
    paste(signif(x / 10^p[[j]], 3), str)
  } else {
    as.character(x)
  }
}
set.seed(1)
```

The `ids` package provides randomly generated ids in a number of
different forms with different readability and sizes.

## Random bytes

The `random_id` function generates random identifiers by generating
`bytes` random bytes and converting to hexadecimal (so each byte
becomes a pair of characters).  Rather than use R's random number
stream we use the `openssl` package here.
``` {r }
ids::random_id()
```

All `ids` functions take `n` as the first argument to be the number
of identifiers generated:
``` {r }
ids::random_id(5)
```

The default here is 16 bytes, each of which has 256 values (so
256^16 = 2^128 = 3.4e38 combinations).  You can make these larger or
smaller with the `bytes` argument:
``` {r }
ids::random_id(5, 8)
```

If `NULL` is provided as `n`, then a generating function is
returned (all ids functions do this):
``` {r }
f <- ids::random_id(NULL, 8)
f
```

This function sets all arguments except for `n`
``` {r }
f()
f(4)
```

## UUIDs

The above look a lot like UUIDs but they are not actually UUIDs.
The `uuid` package provides real UUIDs generated with libuuid, and
the `ids::uuid` function provides an interface to that:
``` {r }
ids::uuid()
```

As above, generate more than one UUID:
``` {r }
ids::uuid(4)
```

Generate time-based UUIDs:
``` {r }
ids::uuid(4, use_time = TRUE)
```

and optionally drop the hyphens:
``` {r }
ids::uuid(5, drop_hyphens = TRUE)
```

## Adjective animal

Generate (somewhat) human readable identifiers by combining one or
more adjectives with an animal name.
``` {r }
ids::adjective_animal()
```

The list of adjectives and animals comes from
[gfycat.com](http://gfycat.com), via
https://github.com/a-type/adjective-adjective-animal

Generate more than one identifier:
``` {r }
ids::adjective_animal(4)
```

Use more than one adjective for very long idenfiers
``` {r }
ids::adjective_animal(4, 3)
```

``` {r echo = FALSE, results = "hide"}
n1 <- length(ids:::gfycat_animals)
n2 <- length(ids:::gfycat_adjectives)
```

There are `r n1` animal names and `r n2` adjectives so each one you
add increases the idenfier space by a factor of `r n2`.  So for 1,
2, and 3 adjectives there are about `r human_no(n1 * n2)`,
`r human_no(n1 * n2^2)` and `r human_no(n1 * n2^3)` possible combinations.

This is a much smaller space than the random identifiers above, but
these are more readable and memorable.

Note that here, the random nunbers are coming from R's random
number stream so are affected by `set.seed()`.

Because some of the animal and adjective names are very long
(e.g. a _quasiextraterritorial hexakosioihexekontahexaphobic
queenalexandrasbirdwingbutterfly_), in order to generate more
readable/memorable identifiers it may be useful to restrict the
length.  Pass `max_len` in to do this.
``` {r }
ids::adjective_animal(4, max_len = 6)
```

A vector of length 2 here can be used to apply to the adjectives
and animal respectively:
``` {r }
ids::adjective_animal(20, max_len = c(5, Inf))
```

Note that this decreases the pool size and so increases the chance
of collisions.

In addition to snake_case, the default, the punctuation between
words can be changed to:

kebab-case:
``` {r }
ids::adjective_animal(1, 2, style = "kebab")
```

dot.case:
``` {r }
ids::adjective_animal(1, 2, style = "dot")
```

camel-case:
``` {r }
ids::adjective_animal(1, 2, style = "camel")
```

PascalCase:
``` {r }
ids::adjective_animal(1, 2, style = "pascal")
```

CONSTANT_CASE (aka SHOUTY_CASE)
``` {r }
ids::adjective_animal(1, 2, style = "constant")
```

or with spaces, lower case:
``` {r }
ids::adjective_animal(1, 2, style = "lower")
```

UPPER CASE
``` {r }
ids::adjective_animal(1, 2, style = "upper")
```

Sentence case
``` {r }
ids::adjective_animal(1, 2, style = "sentence")
```

Title Case
``` {r }
ids::adjective_animal(1, 2, style = "title")
```

Again, pass `n = NULL` here to create a generating function:
``` {r }
aa3 <- ids::adjective_animal(NULL, 3, style = "kebab", max_len = c(6, 8))
```

...which can be used to generate ids on demand.
``` {r }
aa3()
aa3(4)
```

## Random sentences

The `sentence` function creates a sentence style identifier.  This
uses the approach described by Asana on [their
blog](https://blog.asana.com/2011/09/6-sad-squid-snuggle-softly).
This approach encodes 32 bits of information (so 2^32 ~= 4 billion
possibilities) and in theory can be remapped to an integer if you
really wanted to.
``` {r }
ids::sentence()
```

As with `adjective_animal`, the case can be changed:
``` {r }
ids::sentence(2, "dot")
```

If you would rather past tense for the verbs, then pass `past = TRUE`:
``` {r }
ids::sentence(4, past = TRUE)
```

## proquints

"proquints" are an identifier that tries to be information dense
but still human readable and (somewhat) pronounceable; "proquint"
stands for *PRO*-nouncable *QUINT*-uplets.  They are introduced in
https://arxiv.org/html/0901.4016

`ids` can generate proquints:

``` {r }
ids::proquint(10)
```

By default it generates two-word proquints but that can be changed:

``` {r }
ids::proquint(5, 1)
ids::proquint(2, 4)
```

Proquints are formed by alternating
consonant/vowel/consonant/vowel/consonant using a subset of both
(16 consonants and 4 vowels).  This yields 2^16 (65,536)
possibilities per word.  Words are always lower case and always
separated by a hyphen.  So with 4 words there are 2^64 combinations
in 23 characters.

Proquints are also useful in that they can be tranlated with
integers.  The proquint `kapop` has integer value 25258
``` {r }
ids::proquint_to_int("kapop")
ids::int_to_proquint(25258)
```

This makes proquints suitable for creating human-pronouncable
identifers out of things like ip addresses, integer primary keys,
etc.

The function `ids::int_to_proquint_word` will translate between
proquint words and integers (and are vectorised)
``` {r }
w <- ids::int_to_proquint_word(sample(2^16, 10) - 1L)
w
```

and `ids::proquint_word_to_int` does the reverse
``` {r }
ids::proquint_word_to_int(w)
```

whille `ids::proquint_to_int` and `ids::int_to_proquint` allows
translation of multi-word proquints.  Overflow is a real
possibility; the maximum integer representable is only about `r
human_no(.Machine$integer.max)` and the maximum floating point
number of accuracy of 1 is about `r human_no(2 /
.Machine$double.eps)` -- these are big numbers but fairly small
proquints:
``` {r }
ids::int_to_proquint(.Machine$integer.max - 1)
ids::int_to_proquint(2 / .Machine$double.eps)
```

But if you had a 6 word proquint this would not work!
``` {r }
p <- ids::proquint(1, 6)
```

Too big for an integer:
``` {r error = TRUE}
ids::proquint_to_int(p)
```

And too big for an numeric number:
``` {r error = TRUE}
ids::proquint_to_int(p, as = "numeric")
```

To allow this, we use `openssl`'s `bignum` support:
``` {r }
ids::proquint_to_int(p, as = "bignum")
```

This returns a *list* with one bignum (this is required to allow
vectorisation).

## Roll your own identifiers

The `ids` functions can build identifiers in the style of
`adjective_animal` or `sentence`.  It takes as input a list of
strings.  This works particularly well with the `rcorpora` package
which includes lists of strings.

Here is a list of Pokemon names:
``` {r }
pokemon <- tolower(rcorpora::corpora("games/pokemon")$pokemon$name)
length(pokemon)
```

...and here is a list of adjectives
``` {r }
adjectives <- tolower(rcorpora::corpora("words/adjs")$adjs)
length(adjectives)
```

So we have a total pool size of about `r human_no(length(adjectives) *
length(pokemon))`, which is not huge, but it is at least topical.

To generate one identifier:
``` {r }
ids::ids(1, adjectives, pokemon)
```

All the style-changing code is available:
``` {r }
ids::ids(10, adjectives, pokemon, style = "dot")
```

Better would be to wrap this so that the constants are not passed
around the whole time:
``` {r }
adjective_pokemon <- function(n = 1, style = "snake") {
  pokemon <- tolower(rcorpora::corpora("games/pokemon")$pokemon$name)
  adjectives <- tolower(rcorpora::corpora("words/adjs")$adjs)
  ids::ids(n, adjectives, pokemon, style = style)
}

adjective_pokemon(10, "kebab")
```

As a second example we can use the word lists in rcorpora to
generate identifiers in the form `<mood>_<scientist>`, like
"melancholic_darwin".  These are similar to the names of
docker containers.

First the lists of names themselves:
``` {r }
moods <- tolower(rcorpora::corpora("humans/moods")$moods)
scientists <- tolower(rcorpora::corpora("humans/scientists")$scientists)
```

Moods include:
``` {r }
sample(moods, 10)
```

The scientists names contain spaces which is not going to work for
us because `ids` won't correctly translate all internal spaces to
the requested style.
``` {r }
sample(scientists, 10)
```

To hack around this we'll just take the last name from the list and
remove all hyphens:
``` {r }
scientists <- vapply(strsplit(sub("(-|jr\\.$)", "", scientists), " "),
                     tail, character(1), 1)
```

Which gives strings that are just letters (though there are a few
non-ASCII characters here that may cause problems because string
handling is just a big pile of awful)
``` {r }
sample(scientists, 10)
```

With the word lists, create an identifier:
``` {r }
ids::ids(1, moods, scientists)
```

Or pass `NULL` for `n` and create a function:
``` {r }
sci_id <- ids::ids(NULL, moods, scientists, style = "kebab")
```

which takes just the number of identifiers to generate as an argument
``` {r }
sci_id(10)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proquint.R
\name{int_to_proquint}
\alias{int_to_proquint}
\alias{proquint_to_int}
\alias{proquint_word_to_int}
\alias{int_to_proquint_word}
\title{Convert to and from proquints}
\usage{
int_to_proquint(x, use_cache = TRUE)

proquint_to_int(p, as = "numeric", use_cache = TRUE)

proquint_word_to_int(w, use_cache = TRUE, validate = TRUE)

int_to_proquint_word(i, use_cache = TRUE, validate = TRUE)
}
\arguments{
\item{x}{An integer (or integer-like) value to convert to a
proquint}

\item{use_cache}{Because there are relatively few combinations per
word, and because constructing short strings is relatively
expensive in R, it may be useful to cache all 65536 possible
words.  If \code{TRUE} then the first time that this function is
used all words will be cached and the results used - the first
time may take up to ~1/4 of a second and subsequent calls will
be much faster.  The identifiers selected will not change with
this option (i.e., given a particular random seed, changing this
option will not affect the identifiers randomly selected).}

\item{p}{A character vector representing a proquint}

\item{as}{The target data type for conversion from proquint.  The
options are \code{integer}, \code{numeric} and \code{bignum}.
The first two will overflow given sufficiently large input -
this will throw an error (overflow is at
\code{.Machine$integer.max} and \code{2 / .Machine$double.eps - 1}
for \code{integer} and \code{numeric} respectively).  For
\code{bignum} this will return a \emph{list} of \code{bignum}
elements \emph{even if \code{p} is of length 1}.}

\item{w}{A proquint \emph{word} (five letter string)}

\item{validate}{Validate the range of inputs?  Because these
functions are used internally, they can skip input validation.
You can too if you promise to pass sanitised input in.  If
out-of-range values are passed in and validation is disabled the
behaviour is undefined and subject to change.}

\item{i}{An integer representing a single proquint word (in the
range 0:65535)}
}
\description{
Convert to and from proquints.
}
\details{
These functions try to be type safe and predictable about what
they will and will not return.

For \code{proquint_to_int}, because numeric overflow is a
possibility, it is important to consider whether a proquint can be
meaninfully translated into an integer or a numeric and the
functions will throw an error rather than failing in a more
insidious way (promoting the type or returning NA).

\code{proquint_word_to_int} always retuns an integer vector of the
same length as the input.

Missing values are allowed; a missing integer representation of a
proquint will translate as \code{NA_character_} and a missing
proquint will translate as \code{NA_integer_} (if \code{as =
"integer"}), \code{NA_real_}, if \code{as = "numeric"} or as
\code{NULL} (if \code{as = "bignum"}).

Names are always discarded.  Future versions may gain an argument
\code{named} with a default of \code{FALSE}, but that setting to
\code{TRUE} would preseve names.  Let me know if this would be
useful.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/random.R
\name{random_id}
\alias{random_id}
\title{Cryptographically generated random identifiers}
\usage{
random_id(n = 1, bytes = 16, use_openssl = TRUE)
}
\arguments{
\item{n}{number of ids to return.  If \code{NULL}, it instead
returns the generating function}

\item{bytes}{The number of bytes to include for each identifier.
The length of the returned identifiers will be twice this long
with each pair of characters representing a single byte.}

\item{use_openssl}{A logical, indicating if we should use the
openssl for generating the random identifiers.  The openssl
random bytestream is not affected by the state of the R random
number generator (e.g., via \code{\link{set.seed}}) so may not be
suitable for use where reproducibility is important.  The speed
should be very similar for both approaches.}
}
\description{
Random identifiers.  By default this uses the \code{openssl}
package to produce a random set of bytes, and expresses that as a
hex character string.  This does not affect R's random number
stream.
}
\examples{
# Generate a random id:
random_id()

# Generate 10 of them!
random_id(10)

# Different length ids
random_id(bytes = 8)
# (note that the number of characters is twice the number of bytes)

# The ids are not affected by R's RNG state:
set.seed(1)
(id1 <- random_id())
set.seed(1)
(id2 <- random_id())
# The generated identifiers are different, despite the seed being the same:
id1 == id2

# If you need these identifiers to be reproducible, pass use_openssl = FALSE
set.seed(1)
(id1 <- random_id(use_openssl = FALSE))
set.seed(1)
(id2 <- random_id(use_openssl = FALSE))
# This time they are the same:
id1 == id2

# Pass \\code{n = NULL} to generate a function that binds your arguments:
id8 <- random_id(NULL, bytes = 8)
id8(10)
}
\author{
Rich FitzJohn
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sentence.R
\name{sentence}
\alias{sentence}
\title{Sentence style identifiers}
\usage{
sentence(n = 1, style = "snake", past = FALSE)
}
\arguments{
\item{n}{number of ids to return.  If \code{NULL}, it instead
returns the generating function}

\item{style}{Style to join words with.  Can be one of "Pascal",
"camel", "snake", "kebab", "dot", "title", "sentence", "lower",
"upper", and "constant".}

\item{past}{Use the past tense for verbs (e.g., slurped or jogged
rather than slurping or jogging)}
}
\description{
Create a sentence style identifier.  This uses the approach
described by Asana on their blog
\url{https://blog.asana.com/2011/09/6-sad-squid-snuggle-softly/}.
This approach encodes 32 bits of information (so 2^32 ~= 4 billion
possibilities) and in theory can be remapped to an integer if you
really wanted to.
}
\examples{
# Generate an identifier
sentence()

# Generate a bunch
sentence(10)

# As with adjective_animal, use "style" to control punctuation
sentence(style = "Camel")
sentence(style = "dot")
sentence(style = "Title")

# Change the tense of the verb:
set.seed(1)
sentence()
set.seed(1)
sentence(past = TRUE)

# Pass n = NULL to bind arguments to a function
id <- sentence(NULL, past = TRUE, style = "dot")
id()
id(10)
}
\author{
Rich FitzJohn
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adjective_animal.R
\name{adjective_animal}
\alias{adjective_animal}
\title{Ids based on a number of adjectives and an animal}
\usage{
adjective_animal(n = 1, n_adjectives = 1, style = "snake",
  max_len = Inf)
}
\arguments{
\item{n}{number of ids to return.  If \code{NULL}, it instead
returns the generating function}

\item{n_adjectives}{Number of adjectives to prefix the anmial with}

\item{style}{Style to join words with.  Can be one of "Pascal",
"camel", "snake", "kebab", "dot", "title", "sentence", "lower",
"upper", and "constant".}

\item{max_len}{The maximum length of a word part to include (this
may be useful because some of the names are rather long.  This
stops you generating a
\code{hexakosioihexekontahexaphobic_queenalexandrasbirdwingbutterfly}).
A vector of length 2 can be passed in here in which case the
first element will apply to the adjectives (all of them) and the
second element will apply to the animals.}
}
\description{
Ids based on a number of adjectives and an animal
}
\details{
The list of adjectives and animals comes from
\url{https://github.com/a-type/adjective-adjective-animal}, and in
turn from \url{gfycat.com}
}
\examples{
# Generate a random identifier:
adjective_animal()

# Generate a bunch all at once:
adjective_animal(5)

# Control the style of punctuation with the style argument:
adjective_animal(style = "lower")
adjective_animal(style = "CONSTANT")
adjective_animal(style = "camel")
adjective_animal(style = "kebab")

# Control the number of adjectives used
adjective_animal(n_adjectives = 3)

# This can get out of hand quickly though:
adjective_animal(n_adjectives = 7)

# Limit the length of adjectives and animals used:
adjective_animal(10, max_len = 6)

# The lengths can be controlled for adjectives and animals
# separately, with Inf meaning no limit:
adjective_animal(10, max_len = c(6, Inf), n_adjectives = 2)

# Pass n = NULL to bind arguments to a function
id <- adjective_animal(NULL, n_adjectives = 2, style = "dot", max_len = 6)
id()
id(10)
}
\author{
Rich FitzJohn
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proquint.R
\name{proquint}
\alias{proquint}
\title{Generate random proquint identifiers}
\usage{
proquint(n = 1, n_words = 2L, use_cache = TRUE, use_openssl = FALSE)
}
\arguments{
\item{n}{number of ids to return.  If \code{NULL}, it instead
returns the generating function}

\item{n_words}{The number of words for each identifier; each word
has \code{2^16} (65536) possible combinations, a two-word
proquint has \code{2^32} possible combinations and an
\code{k}-word proquint has \code{2^(k * 16)} possible
combinations.}

\item{use_cache}{Because there are relatively few combinations per
word, and because constructing short strings is relatively
expensive in R, it may be useful to cache all 65536 possible
words.  If \code{TRUE} then the first time that this function is
used all words will be cached and the results used - the first
time may take up to ~1/4 of a second and subsequent calls will
be much faster.  The identifiers selected will not change with
this option (i.e., given a particular random seed, changing this
option will not affect the identifiers randomly selected).}

\item{use_openssl}{Use openssl for random number generation, with
the primary effect that the identifiers will not be affected by
R's random seed (at a small speed cost).}
}
\description{
Generate random "proquint" identifiers.  "proquint" stands for
PRO-nouncable QUINT-uplets and were described by Daniel Wilkerson
in \url{https://arxiv.org/html/0901.4016}.  Each "word" takes one
of \eqn{2^16} possibilities.  A four word proquint has a keyspace
of \eqn{10^19} possibilities but takes only 23 characters.
Proquint identifiers can be interchanged with integers (though
this is totally optional); see \code{\link{proquint_to_int}} and
the other functions documented on that page.
}
\details{
In the abstract of their paper, Wilkerson introduces proquints:

"Identifiers (IDs) are pervasive throughout our modern
life. We suggest that these IDs would be easier to manage and
remember if they were easily readable, spellable, and
pronounceable. As a solution to this problem we propose using
PRO-nouncable QUINT-uplets of alternating unambiguous consonants
and vowels: proquints."
}
\examples{
# A single, two word, proquint
proquint()

# Longer identifier:
proquint(n_words = 5)

# More identifiers
proquint(10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/uuid.R
\name{uuid}
\alias{uuid}
\title{Generate UUIDs}
\usage{
uuid(n = 1, drop_hyphens = FALSE, use_time = NA)
}
\arguments{
\item{n}{number of ids to return.  If \code{NULL}, it instead
returns the generating function}

\item{drop_hyphens}{Drop the hyphens from the UUID?}

\item{use_time}{Passed through to \code{UUIDgenerate} as \code{use.time}.}
}
\description{
Generate UUIDs using the uuid package.  This is simply a thin
wrapper around \code{uuid::UUIDgenerate} that matches the interface
in the rest of the ids package.
}
\examples{
# Generate one id
uuid()

# Or a bunch
uuid(10)

# More in the style of random_id()
uuid(drop_hyphens = TRUE)
}
\author{
Rich FitzJohn
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ids.R
\name{ids}
\alias{ids}
\title{Generic id generating function}
\usage{
ids(n, ..., vals = list(...), style = "snake")
}
\arguments{
\item{n}{number of ids to return.  If \code{NULL}, it instead
returns the generating function}

\item{...}{A number of character vectors}

\item{vals}{A list of character vectors, \emph{instead} of \code{...}}

\item{style}{Style to join words with.  Can be one of "Pascal",
"camel", "snake", "kebab", "dot", "title", "sentence", "lower",
"upper", and "constant".}
}
\value{
Either a character vector of length \code{n}, or a
  function of one argument if \code{n} is \code{NULL}
}
\description{
Generic id generating function
}
\examples{
# For an example, please see the vignette
}
\author{
Rich FitzJohn
}
